#!/usr/local/bin/perl
use Getopt::Long;
use File::Find;
use strict;

my (%options,$flag1,$flag2,$flag3,$seq,$nsam,$noutg,$length_seq,$length_,$n,$n1,$no,$i,$j,$jj,$S,$name);
my ($pwo,$pwi,$pwi0);
my (@matrixing,@matrixoutg,@matrixnamei,@matrixnameo,@matrixms,@matrixmso,@positionms,@seqnts,@matrixo,@matrixi,@positionmsrel,@matrixmsMS,@matrixmsoMS);
my ($filenameout,$line,$nt);

#options
GetOptions(
     'in=s'   => \$options{'in'},
     'format=s' => \$options{'format'},
     'outg=s' => \$options{'outg'},
     'div=s'  => \$options{'div'},
     'mis=s'  => \$options{'mis'},
     'gap=s'  => \$options{'gap'},
     'cols=s'  => \$options{'cols'},
     'out=s'  => \$options{'out'},
);

#Usage
if($options{'in'} eq undef) {
   print "\nUsage: \nperl seq2tab.pl -in [inputfile] -format [fasta : nbrf] -outg [name or fragment of the outgroup name] -div [0:no print div; 1:include div] -mis [0:no missing; 1:include missing] -gap [0:no gaps; 1:include gaps] -cols [output length of cols]\n";
   exit;
}
if($options{'div'} eq undef) {$options{'div'} = '0';}
if($options{'format'} eq undef) {$options{'format'} = 'fasta';}
if($options{'mis'} eq undef) {$options{'mis'} = '0';}
if($options{'gap'} eq undef) {$options{'gap'} = '1';}
if($options{'cols'} eq undef) {$options{'cols'} = '1e4';}
if($options{'out'} eq undef) {$options{'out'} = 'tab';}

#open the input pir file
$flag1=0;
$flag2=0;
$flag3=0;
$nsam =0;
$noutg=0;
open(INPUT,"$options{'in'}") or die;
while(<INPUT>) {
    if($options{'format'} eq 'nbrf') {
      if($flag1==2) {
          if($_=~/>DL;/) {$flag1=$flag2=0;}
          else {
              #check if end of sequence
              if($_=~/\*/) {
                  ($line)=($_=~/(.+)\*/);
                  $flag3=1;
              }
              else {($line)=($_=~/(.+)/);}
              $seq = $seq.$line;
              #add sequence to matrix
              if($flag3==1) {
                  $seq =~ s/\s//gm; #eliminate spaces, tabs etc...
                  if($flag2==0) {push @matrixing,$seq;}
                  else {push @matrixoutg,$seq;}
              }
          }
      }
      if($flag1==1) {
          $flag1=2;
          $nsam = $nsam + 1;
      };
      if($_=~/>DL;s*(.+)/) {
          $flag1=1;
          $flag3=0;
          $name = $1;
          $seq = "";
          if($options{'outg'} ne undef) {
              if($_=~/$options{'outg'}/) {
                  $flag2=1;
                  $matrixnameo[$noutg] = $name;
                  $noutg = $noutg + 1;
              }
              else {$matrixnamei[$nsam-$noutg] = $name;}
          }
          else {
              $flag2=0;
              $matrixnamei[$nsam-$noutg] = $name;
          }
      }
    }
    if($options{'format'} eq 'fasta') {
      if($flag1==1) {
         if($_=~/>/) {
            #end of sequence, add sequence to matrix
            $seq =~ s/\s//gm; #eliminate spaces, tabs etc...
            if($flag2==0) {push @matrixing,$seq;}
            else {push @matrixoutg,$seq;}
            $flag1=$flag2=0;
         }
         else {
            ($line)=($_=~/(.+)/);
            $seq = $seq.$line;
         }
      }
      if($_=~/>s*(.+)/) {
          $flag1=1;
          $name = $1;
          $seq = "";
          if($options{'outg'} ne undef) {
               if($_=~/$options{'outg'}/) {
                  $flag2=1;
                  $matrixnameo[$noutg] = $name;
                  $noutg = $noutg + 1;
               }
               else {
                  $matrixnamei[$nsam-$noutg] = $name;
               }
          }
          else {
              $flag2=0;
              $matrixnamei[$nsam-$noutg] = $name;
          }
          $nsam = $nsam + 1;
      }
    }
}
close(INPUT);

if($options{'format'} eq 'fasta'){
   #end of last sequence, add sequence to matrix
   $seq =~ s/\s//gm; #eliminate spaces, tabs etc...
   if($flag2==0) {push @matrixing,$seq;}
   else {push @matrixoutg,$seq;}
}

#checking
#foreach $line (@matrixing) {print "\n$line";}
#foreach $line (@matrixoutg) {print "\n$line";}

$nsam = $nsam - $noutg;
if($nsam < 2) {
    print "Error: at least two individuals are necessary in the population.";
    exit;
}
#check length of sequences
$length_seq = length @matrixing[0];
for($n=1;$n<$nsam;$n++) {
    if(length @matrixing[$n] != $length_seq) {
        print "Error: sequences have not the same length.";
        #print "Remember spaces are not accepted.";
        exit;
    }
}
for($n=0;$n<$noutg;$n++) {
    if(length @matrixoutg[$n] != $length_seq) {
        print "Error: outgroup sequences have not the same length.";
        #print "Remember spaces are not accepted.";
        exit;
    }
}
#change all sequences to uppercase
for($n=0;$n<$nsam;$n++) {
    @matrixing[$n]=~ tr/[a-z]/[A-Z]/;
}
for($n=0;$n<$noutg;$n++) {
   @matrixoutg[$n]=~ tr/[a-z]/[A-Z]/;
}
#check for AGCTN
for($n=1;$n<$nsam;$n++) {
    @matrixing[$n]=~/([GATCSWRYKMN\-\?]+)/;
    $length_ = length $1;
    if($length_ != $length_seq) {
        print 'Error: characters accepted are: AGCTagctNnWSRYMKwsrymk?-';
        exit;
    }
}
for($n=1;$n<$noutg;$n++) {
    @matrixoutg[$n]=~/([GATCSWRYKMN\-\?]+)/;
    $length_ = length $1;
    if($length_ != $length_seq) {
        print 'Error: characters accepted are: AGCTagctNnWSRYMKwsrymk?-';
        exit;
    }
}

#transform sequences in separated nucleotides per sample
for($n=0;$n<$noutg;$n++) {push @matrixo, [split //,@matrixoutg[$n]];}
for($n=0;$n<$nsam;$n++) {push @matrixi, [split //,@matrixing[$n]];}

#checking
#for($n=0;$n<$nsam;$n++)  {for($i=0;$i<$length_seq;$i++) {print "$matrixi[$n][$i]";} print "\n";}
#for($n=0;$n<$noutg;$n++) {for($i=0;$i<$length_seq;$i++) {print "$matrixo[$n][$i]";} print "\n";}

#transform data in a table of polymorphisms
$S=0;
for($i=0;$i<$length_seq;$i++) {
   if($options{'gap'}=='0') {
      $n1=0;
      while($n1 < $nsam && $matrixi[$n1][$i]!~/[\-]/) {$n1++;}
      if($n1 < $nsam) {next;}
      $n1=0;
      while($n1 < $noutg && $matrixo[$n1][$i]!~/[\-]/) {$n1++;}
      if($n1 < $noutg) {next;}
   }
   if($options{'mis'}=='0') {
      $n1=0;
      while($n1 < $nsam && $matrixi[$n1][$i]!~/[N\?]/) {$n1++;}
      if($n1 < $nsam) {next;}
      $n1=0;
      while($n1 < $noutg && $matrixo[$n1][$i]!~/[N\?]/) {$n1++;}
      if($n1 < $noutg) {next;}
   }
   if($options{'div'}=='0') {
      $pwo = 1;
      $pwi = 0;
      $n1=0;
      while($n1 < $nsam && $matrixi[$n1][$i]=~/[N\?]/) {$n1++;} #[N\-\?]
      if($n1 < $nsam) {
         $pwi0 = $matrixi[$n1][$i];
         for($n=$n1+1;$n<$nsam;$n++) {
            if(($matrixi[$n][$i] ne $pwi0 && $matrixi[$n][$i] !~/[N\?]/) || ($matrixi[$n][$i]=~/[SWRYKM]/ || $pwi0=~/[SWRYKM]/)) {
               #variant position sample
               $pwi = 1;
               last;
            }
         }
      }
   }
   else {
      $pwo = 0;
      $pwi = 0;
      $n1=0;
      while($n1 < $nsam && $matrixi[$n1][$i]=~/[N\?]/) {$n1++;}
      $no=0;
      while($no < $noutg && $matrixo[$no][$i]=~/[N\?]/) {$no++;}
      if(($n1 < $nsam) & ($no < $noutg)) {
         $pwi0 = $matrixi[$n1][$i];
         for($n=$n1+1;$n<$nsam;$n++) {
            if((($matrixi[$n][$i] ne $pwi0) && ($matrixi[$n][$i] !~/[N\?]/) && ($pwi0 !~/[N\?]/)) || ($matrixi[$n][$i] =~/[SWRYKM]/ || $pwi0 =~/[SWRYKM]/)) {
               #variant position sample
               $pwi = 1;
               $pwo = 1;
               last;
            }
         }
         if($pwi == 0) {
            for($n=0;$n<$noutg;$n++) {
               if(($matrixo[$n][$i] ne $pwi0 && $matrixo[$n][$i] !~/[N\?]/ && $pwi0 !~/[N\?]/) || ($matrixo[$n][$i]=~/[SWRYKM]/ || $pwi0=~/[SWRYKM]/)) {
                  #variant position sample vs outgroup
                  $pwi = 1;
                  $pwo = 1;
                  last;
               }
            }
            for($n=$no+1;$n<$noutg;$n++) {
               if(($matrixo[$no][$i] ne $matrixo[$n][$i] && $matrixo[$n][$i] !~/[N\?]/ && $matrixo[$no][$i] !~/[N\?]/) || ($matrixo[$n][$i]=~/[SWRYKM]/ || $matrixo[$no][$i]=~/[SWRYKM]/)) {
                  #variant position within outgroup
                  $pwi = 1;
                  $pwo = 1;
                  last;
               }
            }
         }
      }
   }
   #include variants in the matrix.
   if($pwo == 1 && $pwi == 1) {
      #only variants are considered
      $positionms[$S] = $i+1;
      $matrixms[0][$S] = $matrixi[0][$i];      
      for($n=1;$n<$nsam;$n++) {
          if($matrixi[$n][$i] eq $matrixi[0][$i] && $matrixi[0][$i] !~/[N\-\?]/) {$matrixms[$n][$S] ='.';}
          else {$matrixms[$n][$S] = $matrixi[$n][$i];$matrixmsMS[$n][$S] = '1';}
          #$matrixms[$n][$S] = $matrixi[$n][$i];
      }
      for($n=0;$n<$noutg;$n++) {
          if($matrixo[$n][$i] eq $matrixi[0][$i] && $matrixi[0][$i] !~/[N\-\?]/) {$matrixmso[$n][$S] ='.';}
          else {$matrixmso[$n][$S] = $matrixo[$n][$i];}
          #$matrixmso[$n][$S] = $matrixo[$n][$i];
      }

      my $nreal = 0;
      while($matrixi[$nreal][$i] =~/[N\-\?]/ && $nreal < $nsam) {$nreal++;}
      if($nreal == $nsam) {$nreal = 0;}
      $matrixmsMS[$nreal][$S] = '0';
      for($n=0;$n<$nsam;$n++) {
          if($matrixi[$n][$i] eq $matrixi[$nreal][$i]) {$matrixmsMS[$n][$S] = '0';}
          else {$matrixmsMS[$n][$S] = '1';}
      }
      for($n=0;$n<$noutg;$n++) {
          if($matrixo[$n][$i] eq $matrixi[$nreal][$i]) {$matrixmsoMS[$n][$S] = '0';}
          else {$matrixmsoMS[$n][$S] = '1';}
      }
      $S = $S + 1;
   }
}

#open the output file
if($options{'out'} eq "tab") {
($options{'in'})=~/(\w+)/;
if($options{'div'}=='1') {
   if($options{'gap'}=='0') {
      if($options{'mis'}=='0') {
         $filenameout = ">"."$1"."_d".".txt";
      }
      else {
         $filenameout = ">"."$1"."_dm".".txt";
      }
   }
   else {
      if($options{'mis'}=='0') {
         $filenameout = ">"."$1"."_dg".".txt";
      }
      else {
         $filenameout = ">"."$1"."_dmg".".txt";
      }
   }
}
else {
   if($options{'gap'}=='0') {
      if($options{'mis'}=='0') {
         $filenameout = ">"."$1"."_p".".txt";
      }
      else {
         $filenameout = ">"."$1"."_pm".".txt";
      }
   }
   else {
      if($options{'mis'}=='0') {
         $filenameout = ">"."$1"."_pg".".txt";
      }
      else {
         $filenameout = ">"."$1"."_pmg".".txt";
      }
   }
}
open(OUTPUT,$filenameout) or die;

#print table in the output file
print OUTPUT "variants: $S\n";
for($j=0;$j<$S;$j= $j+$options{'cols'}) {
    if($S < $j+$options{'cols'}) {$jj = $S;}
    else {$jj = $j+$options{'cols'};}
    
    print OUTPUT "aligned positions: \n\t ";
    for($i=$j;$i<$jj;$i++) {print OUTPUT "$positionms[$i] ";}
    print OUTPUT "\n";
    for($n=0;$n<$nsam;$n++) {
        print OUTPUT "$matrixnamei[$n]\t"; 
        for($i=$j;$i<$jj;$i++) {print OUTPUT "$matrixms[$n][$i] ";} 
        print OUTPUT "\n";
    }
    for($n=0;$n<$noutg;$n++) {
        print OUTPUT "$matrixnameo[$n]\t"; 
        for($i=$j;$i<$jj;$i++) {print OUTPUT "$matrixmso[$n][$i] ";} 
        print OUTPUT "\n";
    }
    print OUTPUT "\n";
}

#close output file
close(OUTPUT);
}

if($options{'out'} eq "ms") {
($options{'in'})=~/(\w+)/;
if($options{'div'}=='1') {
      if($options{'mis'}=='0') {
         $filenameout = ">"."$1"."_d".".ms.txt";
      }
      else {
         $filenameout = ">"."$1"."_dm".".ms.txt";
      }
}
else {
      if($options{'mis'}=='0') {
         $filenameout = ">"."$1"."_p".".ms.txt";
      }
      else {
         $filenameout = ">"."$1"."_pm".".ms.txt";
      }
}
open(OUTPUT,$filenameout) or die;

#print table in the output file
print OUTPUT "seq2tab.pl -in $options{'in'} -format $options{'format'} -outg $options{'outg'} -div $options{'div'} -mis $options{'mis'} -gap $options{'gap'} -cols $options{'cols'} -out $options{'out'}\n";
print OUTPUT "\n";
print OUTPUT "\n";
print OUTPUT "// \n";
print OUTPUT "segsites: $S\n";
for($j=0;$j<$S;$j= $j+$options{'cols'}) {
    if($S < $j+$options{'cols'}) {$jj = $S;}
    else {$jj = $j+$options{'cols'};}
    
    print OUTPUT "positions: ";
    my $posrel;
    for($i=$j;$i<$jj;$i++) {$positionmsrel[$i]=($positionms[$i]-1)/$length_seq;}
    for($i=$j;$i<$jj;$i++) {print OUTPUT "$positionmsrel[$i] ";}
        print OUTPUT "\n";
    for($n=0;$n<$nsam;$n++) {
        for($i=$j;$i<$jj;$i++) {print OUTPUT "$matrixmsMS[$n][$i]";} 
        print OUTPUT "\n";
    }
    for($n=0;$n<$noutg;$n++) {
        for($i=$j;$i<$jj;$i++) {print OUTPUT "$matrixmsoMS[$n][$i]";} 
        print OUTPUT "\n";
    }
    print OUTPUT "\n";
}

#close output file
close(OUTPUT);
}

